@@ -17,7 +17,7 @@
 class MigratorConstants:
     @staticmethod
     def get_version():
-        return '0.9.1'
+        return '0.10.0'
 
     @staticmethod
     def get_full_name():
@@ -47,6 +47,10 @@ class MigratorConstants:
     def get_default_indent():
         return '    '
 
+    @staticmethod
+    def get_default_data_source():
+        return 'SOURCE TABLE'
+
     @staticmethod
     def get_internal_configuration():
         return {
