@@ -232,7 +232,15 @@ class DatabaseConnector(ABC):
         """
         Migrate a table from source to target database.
         Procedure is used inside a worker thread.
-        Returns number of rows migrated.
+        Returns dictionary migration_stats:
+        {
+            'finished': bool - True if whole migration of this table is fully finished, False if not
+            'rows_migrated': int - number of rows migrated in this chunk
+            'source_table_rows': int - total number of rows in the source table
+            'target_table_rows': int - total number of rows in the target table after this chunk
+            'chunk_number': int - current chunk number
+            'total_chunks': int - total number of chunks for this table
+        }
         """
         pass
 
@@ -601,13 +609,60 @@ class DatabaseConnector(ABC):
         pass
 
     @abstractmethod
-    def get_top10_biggest_tables(self, settings):
+    def get_top_n_tables(self, settings):
         """
         Settings - dictionary with the following keys
             - source_schema: str - schema name of the tables to be checked
-        Returns the top 10 biggest tables in the database.
-        This is used for debugging purposes and for checking compatibility with the migrator.
-        Returns a dictionary: {ordinal_number: {table_name: table_size}} sorted by table_size in descending order.
+        Returns a dictionary with the top N tables in the specified schema.
+        The dictionary contains the following keys:
+            - 'by_rows': dict - top tables by number of rows
+            - 'by_size': dict - top tables by size in bytes
+            - 'by_columns': dict - top tables by number of columns
+            - 'by_indexes': dict - top tables by number of indexes
+            - 'by_constraints': dict - top tables by number of constraints
+        Each of these keys contains a dictionary with structure like this (not all keys are used in all cases):
+        { ordinary_number: {
+            'table_name': table_name,
+            'table_schema': table_schema,
+            'table_size': table_size,  # in bytes
+            'table_rows': table_rows,  # number of rows
+            'table_columns': table_columns,  # number of columns
+            'table_indexes': table_indexes,  # number of indexes
+            'table_constraints': table_constraints,  # number of constraints
+            }
+        """
+        pass
+
+    @abstractmethod
+    def get_top_fk_dependencies(self, settings):
+        """
+        Fetch top foreign key dependencies in the specified schema.
+        settings - dictionary with the following keys
+            - source_schema: str - schema name of the tables to be checked
+        Returns a dictionary with the top foreign key dependencies.
+        Each of these keys contains a dictionary with structure like this:
+        { ordinary_number: {
+            'owner': owner_name,
+            'table_name': table_name,
+            'fk_count': foreign_key_count,
+            'dependencies: list of source tables that have foreign key references to this table
+            }
+        }
+        """
+        pass
+
+    @abstractmethod
+    def target_table_exists(self, target_schema, target_table):
+        """
+        Check if the target table exists in the target database.
+        Returns True if the table exists, False otherwise.
+        """
+        pass
+
+    @abstractmethod
+    def fetch_all_rows(self, query):
+        """
+        Fetch all rows from the database using the provided query.
         """
         pass
 
